"""Generic class to handle Image-like data in a OME-NGFF file."""

from collections.abc import Sequence
from typing import Generic, Literal, TypeVar

import dask.array as da
import numpy as np
import zarr

from ngio.common import (
    Dimensions,
    InterpolationOrder,
    Roi,
    RoiPixels,
    consolidate_pyramid,
)
from ngio.io_pipes import (
    DaskGetter,
    DaskRoiGetter,
    DaskRoiSetter,
    DaskSetter,
    NumpyGetter,
    NumpyRoiGetter,
    NumpyRoiSetter,
    NumpySetter,
    SlicingInputType,
    TransformProtocol,
)
from ngio.ome_zarr_meta import (
    AxesHandler,
    Dataset,
    ImageMetaHandler,
    LabelMetaHandler,
    PixelSize,
)
from ngio.tables import RoiTable
from ngio.utils import NgioFileExistsError, ZarrGroupHandler

_image_handler = TypeVar("_image_handler", ImageMetaHandler, LabelMetaHandler)


class AbstractImage(Generic[_image_handler]):
    """A class to handle a single image (or level) in an OME-Zarr image.

    This class is meant to be subclassed by specific image types.
    """

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: _image_handler,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.

        """
        self._path = path
        self._group_handler = group_handler
        self._meta_handler = meta_handler

        try:
            self._zarr_array = self._group_handler.get_array(self._path)
        except NgioFileExistsError as e:
            raise NgioFileExistsError(f"Could not find the dataset at {path}.") from e

    def __repr__(self) -> str:
        """Return a string representation of the image."""
        return f"Image(path={self.path}, {self.dimensions})"

    @property
    def path(self) -> str:
        """Return the path of the image."""
        return self._path

    @property
    def meta_handler(self) -> _image_handler:
        """Return the metadata."""
        return self._meta_handler

    @property
    def dataset(self) -> Dataset:
        """Return the dataset of the image."""
        return self.meta_handler.meta.get_dataset(path=self.path)

    @property
    def dimensions(self) -> Dimensions:
        """Return the dimensions of the image."""
        return Dimensions(
            shape=self.zarr_array.shape,
            chunks=self.zarr_array.chunks,
            dataset=self.dataset,
        )

    @property
    def pixel_size(self) -> PixelSize:
        """Return the pixel size of the image."""
        return self.dataset.pixel_size

    @property
    def axes_handler(self) -> AxesHandler:
        """Return the axes handler of the image."""
        return self.dataset.axes_handler

    @property
    def axes(self) -> tuple[str, ...]:
        """Return the axes of the image."""
        return self.dimensions.axes

    @property
    def zarr_array(self) -> zarr.Array:
        """Return the Zarr array."""
        return self._zarr_array

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the image."""
        return self.zarr_array.shape

    @property
    def dtype(self) -> str:
        """Return the dtype of the image."""
        return str(self.zarr_array.dtype)

    @property
    def chunks(self) -> tuple[int, ...]:
        """Return the chunks of the image."""
        return self.zarr_array.chunks

    @property
    def is_3d(self) -> bool:
        """Return True if the image is 3D."""
        return self.dimensions.is_3d

    @property
    def is_2d(self) -> bool:
        """Return True if the image is 2D."""
        return self.dimensions.is_2d

    @property
    def is_time_series(self) -> bool:
        """Return True if the image is a time series."""
        return self.dimensions.is_time_series

    @property
    def is_2d_time_series(self) -> bool:
        """Return True if the image is a 2D time series."""
        return self.dimensions.is_2d_time_series

    @property
    def is_3d_time_series(self) -> bool:
        """Return True if the image is a 3D time series."""
        return self.dimensions.is_3d_time_series

    @property
    def is_multi_channels(self) -> bool:
        """Return True if the image is multichannel."""
        return self.dimensions.is_multi_channels

    @property
    def space_unit(self) -> str | None:
        """Return the space unit of the image."""
        return self.axes_handler.space_unit

    @property
    def time_unit(self) -> str | None:
        """Return the time unit of the image."""
        return self.axes_handler.time_unit

    def has_axis(self, axis: str) -> bool:
        """Return True if the image has the given axis."""
        return self.axes_handler.has_axis(axis)

    def _get_as_numpy(
        self,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray:
        """Get the image as a numpy array.

        Args:
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        numpy_getter = NumpyGetter(
            zarr_array=self.zarr_array,
            dimensions=self.dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
        )
        return numpy_getter()

    def _get_roi_as_numpy(
        self,
        roi: Roi | RoiPixels,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray:
        """Get the image as a numpy array for a region of interest.

        Args:
            roi: The region of interest to get the array.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        numpy_roi_getter = NumpyRoiGetter(
            zarr_array=self.zarr_array,
            dimensions=self.dimensions,
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
        )
        return numpy_roi_getter()

    def _get_as_dask(
        self,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> da.Array:
        """Get the image as a dask array.

        Args:
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.
        """
        dask_getter = DaskGetter(
            zarr_array=self.zarr_array,
            dimensions=self.dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
        )
        return dask_getter()

    def _get_roi_as_dask(
        self,
        roi: Roi | RoiPixels,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> da.Array:
        """Get the image as a dask array for a region of interest.

        Args:
            roi: The region of interest to get the array.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.
        """
        roi_dask_getter = DaskRoiGetter(
            zarr_array=self.zarr_array,
            dimensions=self.dimensions,
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_kwargs,
        )
        return roi_dask_getter()

    def _get_array(
        self,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray | da.Array:
        """Get a slice of the image.

        Args:
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            mode: The object type to return.
                Can be "dask", "numpy".
            **slicing_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        if mode == "numpy":
            return self._get_as_numpy(
                axes_order=axes_order, transforms=transforms, **slicing_kwargs
            )
        elif mode == "dask":
            return self._get_as_dask(
                axes_order=axes_order, transforms=transforms, **slicing_kwargs
            )
        else:
            raise ValueError(
                f"Unsupported mode: {mode}. Supported modes are: numpy, dask."
            )

    def _get_roi(
        self,
        roi: Roi | RoiPixels,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        **slice_kwargs: SlicingInputType,
    ) -> np.ndarray | da.Array:
        """Get a slice of the image.

        Args:
            roi: The region of interest to get the array.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            mode: The mode to return the array.
                Can be "dask", "numpy".
            **slice_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        if mode == "numpy":
            return self._get_roi_as_numpy(
                roi=roi, axes_order=axes_order, transforms=transforms, **slice_kwargs
            )
        elif mode == "dask":
            return self._get_roi_as_dask(
                roi=roi, axes_order=axes_order, transforms=transforms, **slice_kwargs
            )
        else:
            raise ValueError(
                f"Unsupported mode: {mode}. Supported modes are: numpy, dask."
            )

    def _set_array(
        self,
        patch: np.ndarray | da.Array,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> None:
        """Set a slice of the image.

        Args:
            patch: The patch to set.
            axes_order: The order of the axes to set the patch.
            transforms: The transforms to apply to the patch.
            **slicing_kwargs: The slices to set the patch.

        """
        if isinstance(patch, np.ndarray):
            numpy_setter = NumpySetter(
                zarr_array=self.zarr_array,
                dimensions=self.dimensions,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
            )
            numpy_setter(patch)

        elif isinstance(patch, da.Array):
            dask_setter = DaskSetter(
                zarr_array=self.zarr_array,
                dimensions=self.dimensions,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
            )
            dask_setter(patch)
        else:
            raise TypeError(
                f"Unsupported patch type: {type(patch)}. "
                "Supported types are: "
                "numpy.ndarray, dask.array.Array."
            )

    def _set_roi(
        self,
        roi: Roi | RoiPixels,
        patch: np.ndarray | da.Array,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> None:
        """Set a slice of the image.

        Args:
            roi: The region of interest to set the patch.
            patch: The patch to set.
            axes_order: The order of the axes to set the patch.
            transforms: The transforms to apply to the patch.
            **slicing_kwargs: The slices to set the patch.

        """
        if isinstance(patch, np.ndarray):
            roi_numpy_setter = NumpyRoiSetter(
                zarr_array=self.zarr_array,
                dimensions=self.dimensions,
                roi=roi,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
            )
            roi_numpy_setter(patch)

        elif isinstance(patch, da.Array):
            roi_dask_setter = DaskRoiSetter(
                zarr_array=self.zarr_array,
                dimensions=self.dimensions,
                roi=roi,
                axes_order=axes_order,
                transforms=transforms,
                slicing_dict=slicing_kwargs,
            )
            roi_dask_setter(patch)
        else:
            raise TypeError(
                f"Unsupported patch type: {type(patch)}. "
                "Supported types are: "
                "numpy.ndarray, dask.array.Array."
            )

    def _consolidate(
        self,
        order: InterpolationOrder = "linear",
        mode: Literal["dask", "numpy", "coarsen"] = "dask",
    ) -> None:
        """Consolidate the image on disk.

        Args:
            order: The order of the consolidation.
            mode: The mode of the consolidation.
        """
        consolidate_image(image=self, order=order, mode=mode)

    def roi(self, name: str | None = "image") -> Roi:
        """Return the ROI covering the entire image."""
        dim_x = self.dimensions.get("x")
        dim_y = self.dimensions.get("y")
        assert dim_x is not None and dim_y is not None
        dim_z = self.dimensions.get("z")
        z = None if dim_z is None else 0
        dim_t = self.dimensions.get("t")
        t = None if dim_t is None else 0
        roi_px = RoiPixels(
            name=name,
            x=0,
            y=0,
            z=z,
            t=t,
            x_length=dim_x,
            y_length=dim_y,
            z_length=dim_z,
            t_length=dim_t,
        )
        return roi_px.to_roi(pixel_size=self.pixel_size)

    def build_image_roi_table(self, name: str | None = "image") -> RoiTable:
        """Build the ROI table containing the ROI covering the entire image."""
        return RoiTable(rois=[self.roi(name=name)])

    def require_dimensions_match(
        self,
        other: "AbstractImage",
        allow_singleton: bool = False,
    ) -> None:
        """Assert that two images have matching spatial dimensions.

        Args:
            other: The other image to compare to.
            allow_singleton: If True, allow singleton dimensions to be
                compatible with non-singleton dimensions.

        Raises:
            NgioValueError: If the images do not have compatible dimensions.
        """
        self.dimensions.require_dimensions_match(
            other.dimensions, allow_singleton=allow_singleton
        )

    def check_if_dimensions_match(
        self,
        other: "AbstractImage",
        allow_singleton: bool = False,
    ) -> bool:
        """Check if two images have matching spatial dimensions.

        Args:
            other: The other image to compare to.
            allow_singleton: If True, allow singleton dimensions to be
                compatible with non-singleton dimensions.

        Returns:
            bool: True if the images have matching dimensions, False otherwise.
        """
        return self.dimensions.check_if_dimensions_match(
            other.dimensions, allow_singleton=allow_singleton
        )

    def require_axes_match(
        self,
        other: "AbstractImage",
    ) -> None:
        """Assert that two images have compatible axes.

        Args:
            other: The other image to compare to.

        Raises:
            NgioValueError: If the images do not have compatible axes.
        """
        self.dimensions.require_axes_match(other.dimensions)

    def check_if_axes_match(
        self,
        other: "AbstractImage",
    ) -> bool:
        """Check if two images have compatible axes.

        Args:
            other: The other image to compare to.

        Returns:
            bool: True if the images have compatible axes, False otherwise.

        """
        return self.dimensions.check_if_axes_match(other.dimensions)

    def require_rescalable(
        self,
        other: "AbstractImage",
    ) -> None:
        """Assert that two images can be rescaled to each other.

        For this to be true, the images must have the same axes, and
        the pixel sizes must be compatible (i.e. one can be scaled to the other).

        Args:
            other: The other image to compare to.

        Raises:
            NgioValueError: If the images cannot be scaled to each other.
        """
        self.dimensions.require_rescalable(other.dimensions)

    def check_if_rescalable(
        self,
        other: "AbstractImage",
    ) -> bool:
        """Check if two images can be rescaled to each other.

        For this to be true, the images must have the same axes, and
        the pixel sizes must be compatible (i.e. one can be scaled to the other).

        Args:
            other: The other image to compare to.

        Returns:
            bool: True if the images can be rescaled to each other, False otherwise.
        """
        return self.dimensions.check_if_rescalable(other.dimensions)


def consolidate_image(
    image: AbstractImage,
    order: InterpolationOrder = "linear",
    mode: Literal["dask", "numpy", "coarsen"] = "dask",
) -> None:
    """Consolidate the image on disk."""
    target_paths = image._meta_handler.meta.paths
    targets = [
        image._group_handler.get_array(path)
        for path in target_paths
        if path != image.path
    ]
    consolidate_pyramid(
        source=image.zarr_array, targets=targets, order=order, mode=mode
    )



from collections.abc import Iterable, Iterator
from itertools import product
from typing import TypeAlias, TypeVar

from ngio.utils import NgioValueError, ngio_logger

T = TypeVar("T")

##############################################################
#
# Check slice overlaps
#
##############################################################


def _pairs_stream(iterable: Iterable[T]) -> Iterator[tuple[T, T]]:
    # Same as combinations but yields pairs as soon as they are generated
    seen: list[T] = []
    for a in iterable:
        for b in seen:
            yield b, a
        seen.append(a)


SlicingType: TypeAlias = slice | list[int] | int


def check_elem_intersection(s1: SlicingType, s2: SlicingType) -> bool:
    """Compare if two SlicingType elements intersect.

    If they are a slice, check if they overlap.
    If they are integers, check if they are equal.
    If they are lists, check if they have any common elements.
    """
    if not isinstance(s1, type(s2)):
        raise NgioValueError(
            f"Slices must be of the same type. Got {type(s1)} and {type(s2)}"
        )

    if isinstance(s1, slice) and isinstance(s2, slice):
        # Handle slice objects
        start1, stop1, step1 = s1.start or 0, s1.stop or float("inf"), s1.step or 1
        start2, stop2, step2 = s2.start or 0, s2.stop or float("inf"), s2.step or 1

        if step1 is not None and step2 != 1:
            raise NotImplementedError(
                "Intersection for slices with step != 1 is not implemented"
            )

        if step2 is not None and step1 != 1:
            raise NotImplementedError(
                "Intersection for slices with step != 1 is not implemented"
            )

        return not (stop1 <= start2 or stop2 <= start1)
    elif isinstance(s1, int) and isinstance(s2, int):
        # Handle integer indices
        return s1 == s2
    elif isinstance(s1, list) and isinstance(s2, list):
        if set(s1) & set(s2):
            return True
        return False
    else:
        raise TypeError("Unsupported slice type")


def check_slicing_tuple_intersection(
    s1: tuple[SlicingType, ...], s2: tuple[SlicingType, ...]
) -> bool:
    """For a tuple of SlicingType, check if all elements intersect."""
    if len(s1) != len(s2):
        raise NgioValueError("Slices must have the same length")
    return all(check_elem_intersection(a, b) for a, b in zip(s1, s2, strict=True))


def check_if_regions_overlap(slices: Iterable[tuple[SlicingType, ...]]) -> bool:
    """Check for overlaps in a list of slicing tuples using brute-force method.

    This is O(n^2) and not efficient for large lists.
    Returns True if any overlaps are found.
    """
    for it, (si, sj) in enumerate(_pairs_stream(slices)):
        overalap = check_slicing_tuple_intersection(si, sj)
        if overalap:
            return True

        if it == 10_000:
            ngio_logger.warning(
                "Performance Warning check_for_overlaps is O(n^2) and may be slow for "
                "large numbers of regions."
            )
    return False


##############################################################
#
# Check chunk overlaps
#
##############################################################


def _normalize_slice(slc: slice, size: int) -> tuple[int, int]:
    if slc.step not in (None, 1):
        raise NgioValueError(f"Only step=1 slices supported, got step={slc.step}")
    start = 0 if slc.start is None else slc.start
    stop = size if slc.stop is None else slc.stop
    if start < 0 or stop < 0:
        raise NgioValueError("Negative slice bounds are not supported")
    # clamp to [0, size]
    start = min(start, size)
    stop = min(stop, size)
    if start > stop:
        # empty selection
        return (0, 0)
    return start, stop


def _chunk_indices_for_axis(sel: SlicingType, size: int, csize: int) -> list[int]:
    """From a selection for a single axis, return the list chunk indices touched."""
    if isinstance(sel, slice):
        start, stop = _normalize_slice(sel, size)
        if start >= stop:  # empty
            return []
        first = start // csize
        last = (stop - 1) // csize
        return list(range(first, last + 1))

    if isinstance(sel, int):
        if sel < 0 or sel >= size:
            raise IndexError(f"index {sel} out of bounds for axis of size {size}")
        return [sel // csize]

    if isinstance(sel, list):
        if not sel:
            return []
        chunks_hit = {}
        for v in sel:
            if not isinstance(v, int):
                raise TypeError("Only integers allowed inside tuple selections")
            if v < 0 or v >= size:
                raise IndexError(f"index {v} out of bounds for axis of size {size}")
            chunks_hit[v // csize] = None
        return sorted(chunks_hit.keys())

    raise TypeError(f"Unsupported index type: {type(sel)!r}")


def compute_slice_chunks(
    shape: tuple[int, ...],
    chunks: tuple[int, ...],
    slicing_tuple: tuple[SlicingType, ...],
) -> set[tuple[int, ...]]:
    """Compute the set of chunk coordinates touched by `slicing_tuple`.

    Args:
        shape: overall array shape (s1, s2, ...)
        chunks: chunk shape (c1, c2, ...)
        slicing_tuple: tuple of slices, ints, or tuples of ints
    """
    if len(slicing_tuple) != len(shape):
        raise NgioValueError(
            f"key must have {len(shape)} items, got {len(slicing_tuple)}"
        )

    per_axis_chunks: list[list[int]] = [
        _chunk_indices_for_axis(sel, size, csize)
        for sel, size, csize in zip(slicing_tuple, shape, chunks, strict=True)
    ]

    # If any axis yields no chunks, the overall selection is empty.
    if any(len(ax) == 0 for ax in per_axis_chunks):
        return set()

    return {tuple(idx) for idx in product(*per_axis_chunks)}


def check_if_chunks_overlap(
    slices: Iterable[tuple[SlicingType, ...]],
    shape: tuple[int, ...],
    chunks: tuple[int, ...],
) -> bool:
    """Check for overlaps in a list of slicing tuples using brute-force method.

    This is O(n^2) and not efficient for large lists.
    Returns True if any overlaps are found.
    """
    slices_chunks = (compute_slice_chunks(shape, chunks, si) for si in slices)
    for it, (si, sj) in enumerate(_pairs_stream(slices_chunks)):
        if si & sj:
            return True
        if it == 10_000:
            ngio_logger.warning(
                "Performance Warning check_for_chunks_overlaps is O(n^2) and may be "
                "slow for large numbers of regions."
            )
    return False



from collections.abc import Sequence
from typing import Protocol, TypeVar

import zarr

from ngio.common._roi import Roi, RoiPixels
from ngio.io_pipes._ops_axes import AxesOps
from ngio.io_pipes._ops_slices import SlicingOps
from ngio.io_pipes._ops_transforms import TransformProtocol

GetterDataType = TypeVar("GetterDataType", covariant=True)
SetterDataType = TypeVar("SetterDataType", contravariant=True)


class DataGetterProtocol(Protocol[GetterDataType]):
    @property
    def zarr_array(self) -> zarr.Array: ...

    @property
    def slicing_ops(self) -> SlicingOps: ...

    @property
    def axes_ops(self) -> AxesOps: ...

    @property
    def transforms(self) -> Sequence[TransformProtocol] | None: ...

    @property
    def roi(self) -> Roi | RoiPixels: ...

    def __call__(self) -> GetterDataType:
        return self.get()

    def get(self) -> GetterDataType: ...


class DataSetterProtocol(Protocol[SetterDataType]):
    @property
    def zarr_array(self) -> zarr.Array: ...

    @property
    def slicing_ops(self) -> SlicingOps: ...

    @property
    def axes_ops(self) -> AxesOps: ...

    @property
    def transforms(self) -> Sequence[TransformProtocol] | None: ...

    @property
    def roi(self) -> Roi | RoiPixels: ...

    def __call__(self, patch: SetterDataType) -> None:
        return self.set(patch)

    def set(self, patch: SetterDataType) -> None: ...



from anndata import AnnData
from anndata._settings import settings
from pandas import DataFrame
from polars import DataFrame as PolarsDataFrame
from polars import LazyFrame

from ngio.tables.backends._abstract_backend import AbstractTableBackend
from ngio.tables.backends._anndata_utils import (
    custom_anndata_read_zarr,
)
from ngio.tables.backends._utils import (
    convert_pandas_to_anndata,
    convert_polars_to_anndata,
    normalize_anndata,
)
from ngio.utils import NgioValueError


class AnnDataBackend(AbstractTableBackend):
    """A class to load and write tables from/to an AnnData object."""

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend."""
        return "anndata"

    @staticmethod
    def implements_anndata() -> bool:
        """Check if the backend implements the anndata protocol."""
        return True

    @staticmethod
    def implements_pandas() -> bool:
        """Whether the handler implements the dataframe protocol."""
        return True

    @staticmethod
    def implements_polars() -> bool:
        """Whether the handler implements the polars protocol."""
        return True

    def load_as_anndata(self) -> AnnData:
        """Load the table as an AnnData object."""
        settings.zarr_write_format = self._group_handler.zarr_format
        anndata = custom_anndata_read_zarr(self._group_handler._group)
        anndata = normalize_anndata(anndata, index_key=self.index_key)
        return anndata

    def load(self) -> AnnData:
        """Load the table as an AnnData object."""
        return self.load_as_anndata()

    def write_from_anndata(self, table: AnnData) -> None:
        """Serialize the table from an AnnData object."""
        full_url = self._group_handler.full_url
        if full_url is None:
            raise NgioValueError(
                f"Ngio does not support writing file from a "
                f"store of type {type(self._group_handler)}. "
                "Please make sure to use a compatible "
                "store like a zarr.DirectoryStore."
            )
        settings.zarr_write_format = self._group_handler.zarr_format
        table.write_zarr(full_url)

    def write_from_pandas(self, table: DataFrame) -> None:
        """Serialize the table from a pandas DataFrame."""
        anndata = convert_pandas_to_anndata(
            table,
            index_key=self.index_key,
        )
        self.write_from_anndata(anndata)

    def write_from_polars(self, table: PolarsDataFrame | LazyFrame) -> None:
        """Consolidate the metadata in the store."""
        anndata = convert_polars_to_anndata(
            table,
            index_key=self.index_key,
        )
        self.write_from_anndata(anndata)


class AnnDataBackendV1(AnnDataBackend):
    """A wrapper for the AnnData backend that for backwards compatibility."""

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend."""
        return "anndata_v1"



from ngio import Roi, RoiPixels
from ngio.images._abstract_image import AbstractImage


def rois_product(rois_a: list[Roi], rois_b: list[Roi]) -> list[Roi]:
    """Compute the product of two sets of ROIs."""
    rois_product = []
    for roi_a in rois_a:
        for roi_b in rois_b:
            intersection = roi_a.intersection(roi_b)
            if intersection:
                rois_product.append(intersection)
    return rois_product


def grid(
    rois: list[Roi],
    ref_image: AbstractImage,
    size_x: int | None = None,
    size_y: int | None = None,
    size_z: int | None = None,
    size_t: int | None = None,
    stride_x: int | None = None,
    stride_y: int | None = None,
    stride_z: int | None = None,
    stride_t: int | None = None,
    base_name: str | None = None,
) -> list[Roi]:
    """This method is a placeholder for creating a regular grid of ROIs."""
    t_dim = ref_image.dimensions.get("t", default=1)
    z_dim = ref_image.dimensions.get("z", default=1)
    y_dim = ref_image.dimensions.get("y", default=1)
    x_dim = ref_image.dimensions.get("x", default=1)

    size_t = size_t if size_t is not None else t_dim
    size_z = size_z if size_z is not None else z_dim
    size_y = size_y if size_y is not None else y_dim
    size_x = size_x if size_x is not None else x_dim

    stride_t = stride_t if stride_t is not None else size_t
    stride_z = stride_z if stride_z is not None else size_z
    stride_y = stride_y if stride_y is not None else size_y
    stride_x = stride_x if stride_x is not None else size_x

    # Here we would create a grid of ROIs based on the specified parameters.
    new_rois = []
    for t in range(0, t_dim, stride_t):
        for z in range(0, z_dim, stride_z):
            for y in range(0, y_dim, stride_y):
                for x in range(0, x_dim, stride_x):
                    roi = RoiPixels(
                        name=base_name,
                        x=x,
                        y=y,
                        z=z,
                        t=t,
                        x_length=size_x,
                        y_length=size_y,
                        z_length=size_z,
                        t_length=size_t,
                    )
                    new_rois.append(roi.to_roi(pixel_size=ref_image.pixel_size))

    return rois_product(rois, new_rois)


def by_yx(rois: list[Roi], ref_image: AbstractImage) -> list[Roi]:
    """Return a new iterator that iterates over ROIs by YX coordinates."""
    return grid(
        rois=rois,
        ref_image=ref_image,
        size_z=1,
        stride_z=1,
        size_t=1,
        stride_t=1,
    )


def by_zyx(rois: list[Roi], ref_image: AbstractImage, strict: bool = True) -> list[Roi]:
    """Return a new iterator that iterates over ROIs by ZYX coordinates."""
    if strict and not ref_image.is_3d:
        raise ValueError(
            "Reference Input image must be 3D to iterate by ZXY coordinates. "
            f"Current dimensions: {ref_image.dimensions}"
        )
    return grid(
        rois=rois,
        ref_image=ref_image,
        size_t=1,
        stride_t=1,
    )


def by_chunks(
    rois: list[Roi],
    ref_image: AbstractImage,
    overlap_xy: int = 0,
    overlap_z: int = 0,
    overlap_t: int = 0,
) -> list[Roi]:
    """This method is a placeholder for chunked processing."""
    chunk_size = ref_image.chunks
    t_axis = ref_image.axes_handler.get_index("t")
    z_axis = ref_image.axes_handler.get_index("z")
    y_axis = ref_image.axes_handler.get_index("y")
    x_axis = ref_image.axes_handler.get_index("x")

    size_x = chunk_size[x_axis] if x_axis is not None else None
    size_y = chunk_size[y_axis] if y_axis is not None else None
    size_z = chunk_size[z_axis] if z_axis is not None else None
    size_t = chunk_size[t_axis] if t_axis is not None else None
    stride_x = size_x - overlap_xy if size_x is not None else None
    stride_y = size_y - overlap_xy if size_y is not None else None
    stride_z = size_z - overlap_z if size_z is not None else None
    stride_t = size_t - overlap_t if size_t is not None else None
    return grid(
        rois=rois,
        ref_image=ref_image,
        size_x=size_x,
        size_y=size_y,
        size_z=size_z,
        size_t=size_t,
        stride_x=stride_x,
        stride_y=stride_y,
        stride_z=stride_z,
        stride_t=stride_t,
    )



from pathlib import Path

from ngio import Roi, create_empty_ome_zarr
from ngio.transforms import ZoomTransform


def test_zoom_from_dimensions(tmp_path: Path):
    full_res_img = create_empty_ome_zarr(
        store=tmp_path / "original.zarr",
        shape=(101, 101),
        axes_names="yx",
        xy_pixelsize=1.0,
    ).get_image()

    img = create_empty_ome_zarr(
        store=tmp_path / "img.zarr",
        shape=(25, 25),
        axes_names="yx",
        xy_pixelsize=4.0,
    ).get_image()

    zoom = ZoomTransform(
        input_image=img,
        target_image=full_res_img,
        order="nearest",
    )
    roi = Roi(name=None, x=0, y=0, x_length=21, y_length=21)

    full_res_data = full_res_img.get_roi_as_numpy(roi=roi)
    rescaled_data = img.get_roi_as_numpy(roi=roi, transforms=[zoom])
    assert full_res_data.shape == rescaled_data.shape, "Failed inbound test"
    try:
        img.set_roi(patch=rescaled_data, roi=roi, transforms=[zoom])
    except Exception as e:
        raise AssertionError(f"Failed inbound test: {e}") from e

    roi = Roi(name=None, x=80, y=80, x_length=21, y_length=21)

    full_res_data = full_res_img.get_roi_as_numpy(roi=roi)
    rescaled_data = img.get_roi_as_numpy(roi=roi, transforms=[zoom])
    assert full_res_data.shape == rescaled_data.shape, "Failed outbound test"
    try:
        img.set_roi(patch=rescaled_data, roi=roi, transforms=[zoom])
    except Exception as e:
        raise AssertionError(f"Failed outbound test: {e}") from e



import pandas as pd
import polars as pl

from ngio.tables.backends._non_zarr_backends import NonZarrBaseBackend


def write_lf_to_csv(path: str, table: pl.DataFrame) -> None:
    """Write a polars DataFrame to a CSV file."""
    table.write_csv(path)


def write_df_to_csv(path: str, table: pd.DataFrame) -> None:
    """Write a pandas DataFrame to a CSV file."""
    table.to_csv(path, index=False)


class CsvTableBackend(NonZarrBaseBackend):
    """A class to load and write small tables in CSV format."""

    def __init__(
        self,
    ):
        """Initialize the CsvTableBackend."""
        super().__init__(
            lf_reader=pl.scan_csv,
            df_reader=pd.read_csv,
            lf_writer=write_lf_to_csv,
            df_writer=write_df_to_csv,
            table_name="table.csv",
        )

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend."""
        return "csv"



from collections.abc import Sequence
from enum import Enum

import dask.array as da
import numpy as np

from ngio.utils import NgioValueError, ngio_logger


class Action(str, Enum):
    NONE = "none"
    PAD = "pad"
    TRIM = "trim"
    RESCALING = "rescaling"


def _compute_pad_widths(
    array_shape: tuple[int, ...],
    actions: list[Action],
    target_shape: tuple[int, ...],
) -> tuple[tuple[int, int], ...]:
    pad_def = []
    for act, s, ts in zip(actions, array_shape, target_shape, strict=True):
        if act == Action.PAD:
            total_pad = ts - s
            before = total_pad // 2
            after = total_pad - before
            pad_def.append((before, after))
        else:
            pad_def.append((0, 0))
    ngio_logger.warning(
        f"Images have a different shape ({array_shape} vs {target_shape}). "
        f"Resolving by padding: {pad_def}",
        stacklevel=2,
    )
    return tuple(pad_def)


def _numpy_pad(
    array: np.ndarray,
    actions: list[Action],
    target_shape: tuple[int, ...],
    pad_mode: str = "constant",
    constant_values: int | float = 0,
) -> np.ndarray:
    if all(act != Action.PAD for act in actions):
        return array
    pad_widths = _compute_pad_widths(array.shape, actions, target_shape)
    return np.pad(array, pad_widths, mode=pad_mode, constant_values=constant_values)  # type: ignore


def _dask_pad(
    array: da.Array,
    actions: list[Action],
    target_shape: tuple[int, ...],
    pad_mode: str = "constant",
    constant_values: int | float = 0,
) -> da.Array:
    if all(act != Action.PAD for act in actions):
        return array
    shape = tuple(int(s) for s in array.shape)
    pad_widths = _compute_pad_widths(shape, actions, target_shape)
    return da.pad(array, pad_widths, mode=pad_mode, constant_values=constant_values)


def _compute_trim_slices(
    array_shape: tuple[int, ...],
    actions: list[Action],
    target_shape: tuple[int, ...],
) -> tuple[slice, ...]:
    slices = []
    for act, s, ts in zip(actions, array_shape, target_shape, strict=True):
        if act == Action.TRIM:
            slices.append(slice(0, ts))
        else:
            slices.append(slice(0, s))

    ngio_logger.warning(
        f"Images have a different shape ({array_shape} vs {target_shape}). "
        f"Resolving by trimming: {slices}",
        stacklevel=2,
    )
    return tuple(slices)


def _numpy_trim(
    array: np.ndarray, actions: list[Action], target_shape: tuple[int, ...]
) -> np.ndarray:
    if all(act != Action.TRIM for act in actions):
        return array
    slices = _compute_trim_slices(array.shape, actions, target_shape)
    return array[tuple(slices)]


def _dask_trim(
    array: da.Array, actions: list[Action], target_shape: tuple[int, ...]
) -> da.Array:
    if all(act != Action.TRIM for act in actions):
        return array
    shape = tuple(int(s) for s in array.shape)
    slices = _compute_trim_slices(shape, actions, target_shape)
    return array[tuple(slices)]


def _compute_rescaling_shape(
    array_shape: tuple[int, ...],
    actions: list[Action],
    target_shape: tuple[int, ...],
) -> tuple[int, ...]:
    rescaling_shape = []
    factor = []
    for act, s, ts in zip(actions, array_shape, target_shape, strict=True):
        if act == Action.RESCALING:
            rescaling_shape.append(ts)
            factor.append(ts / s)
        else:
            rescaling_shape.append(s)
            factor.append(1.0)

    ngio_logger.warning(
        f"Images have a different shape ({array_shape} vs {target_shape}). "
        f"Resolving by scaling with factors {factor}.",
        stacklevel=2,
    )
    return tuple(rescaling_shape)


def _numpy_rescaling(
    array: np.ndarray, actions: list[Action], target_shape: tuple[int, ...]
) -> np.ndarray:
    if all(act != Action.RESCALING for act in actions):
        return array
    from ngio.common._zoom import numpy_zoom

    rescaling_shape = _compute_rescaling_shape(array.shape, actions, target_shape)
    return numpy_zoom(source_array=array, target_shape=rescaling_shape, order="nearest")


def _dask_rescaling(
    array: da.Array, actions: list[Action], target_shape: tuple[int, ...]
) -> da.Array:
    if all(act != Action.RESCALING for act in actions):
        return array
    from ngio.common._zoom import dask_zoom

    shape = tuple(int(s) for s in array.shape)
    rescaling_shape = _compute_rescaling_shape(shape, actions, target_shape)
    return dask_zoom(source_array=array, target_shape=rescaling_shape, order="nearest")


def _check_axes(array_shape, reference_shape, array_axes, reference_axes):
    if len(array_shape) != len(array_axes):
        raise NgioValueError(
            f"Array shape {array_shape} and reference axes {array_axes} "
            "must have the same number of dimensions."
        )
    if len(reference_shape) != len(reference_axes):
        raise NgioValueError(
            f"Reference shape {reference_shape} and reference axes {reference_axes} "
            "must have the same number of dimensions."
        )

    # Check if the array axes are a subset of the target axes
    diff = set(array_axes) - set(reference_axes)
    if diff:
        raise NgioValueError(
            f"Array axes {array_axes} are not a subset "
            f"of reference axes {reference_axes}"
        )

    # Array must be smaller or equal in number of dimensions
    if len(array_axes) > len(reference_axes):
        raise NgioValueError(
            f"Array has more dimensions ({len(array_axes)}) "
            f"than reference ({len(reference_axes)}). "
            "Cannot match shapes if the array has more dimensions."
        )


def _compute_reshape_and_actions(
    array_shape: tuple[int, ...],
    reference_shape: tuple[int, ...],
    array_axes: list[str],
    reference_axes: list[str],
    tolerance: int = 1,
    allow_rescaling: bool = True,
) -> tuple[tuple[int, ...], list[Action]]:
    # Reshape array to match reference shape
    # And determine actions to be taken
    # to match the shapes
    reshape_tuple = []
    actions = []
    errors = []
    left_pointer = 0
    for ref_ax, ref_shape in zip(reference_axes, reference_shape, strict=True):
        if ref_ax not in array_axes:
            reshape_tuple.append(1)
            actions.append(Action.NONE)
        elif ref_ax == array_axes[left_pointer]:
            s2 = array_shape[left_pointer]
            reshape_tuple.append(s2)
            left_pointer += 1

            if s2 == ref_shape or s2 == 1:
                actions.append(Action.NONE)
            elif s2 < ref_shape:
                if (ref_shape - s2) <= tolerance:
                    actions.append(Action.PAD)
                elif allow_rescaling:
                    actions.append(Action.RESCALING)
                else:
                    errors.append(
                        f"Cannot pad axis={ref_ax}:{s2}->{ref_shape} "
                        "because shape difference is outside tolerance "
                        f"{tolerance}."
                    )
            elif s2 > ref_shape:
                if (s2 - ref_shape) <= tolerance:
                    actions.append(Action.TRIM)
                elif allow_rescaling:
                    actions.append(Action.RESCALING)
                else:
                    errors.append(
                        f"Cannot trim axis={ref_ax}:{s2}->{ref_shape} "
                        "because shape difference is outside tolerance "
                        f"{tolerance}."
                    )
            else:
                raise RuntimeError("Unreachable code reached.")
        else:
            raise NgioValueError(
                f"Axes order mismatch {array_axes} -> {reference_axes}. "
                "Cannot match shapes if the order is different."
            )
    if errors:
        raise NgioValueError(
            "Array shape cannot be matched to reference shape:\n\n".join(errors)
        )
    return tuple(reshape_tuple), actions


def numpy_match_shape(
    array: np.ndarray,
    reference_shape: tuple[int, ...],
    array_axes: Sequence[str],
    reference_axes: Sequence[str],
    tolerance: int = 1,
    pad_mode: str = "constant",
    pad_values: int | float = 0,
    allow_rescaling: bool = True,
):
    """Match the shape of a numpy array to a reference shape.

    This function will reshape, pad, trim and broadcast the input array
    to match the reference shape. If the shapes cannot be matched within
    the specified tolerance, an error is raised.

    The reference axes must be a superset of the array axes, and the order
    of the axes must be the same.

    Args:
        array (np.ndarray): The input array to be reshaped.
        reference_shape (tuple[int, ...]): The target shape to match.
        array_axes (Sequence[str]): The axes names of the input array.
        reference_axes (Sequence[str]): The axes names of the reference shape.
        tolerance (int): The maximum number of pixels by which dimensions
            can differ when matching shapes.
        allow_broadcast (bool): If True, allow broadcasting new dimensions to
            match the reference shape. If False, single-dimension axes will
            be left as is.
        pad_mode (str): The mode to use for padding. See numpy.pad for options.
        pad_values (int | float): The constant value to use for padding if
            pad_mode is 'constant'.
        allow_rescaling (bool): If True, when the array differs more than the
            tolerance, it will be rescalingd to the reference shape. If False,
            an error will be raised.
    """
    _check_axes(
        array_shape=array.shape,
        reference_shape=reference_shape,
        array_axes=array_axes,
        reference_axes=reference_axes,
    )
    if array.shape == reference_shape:
        # Shapes already match
        return array

    array_axes = list(array_axes)
    reference_axes = list(reference_axes)

    reshape_tuple, actions = _compute_reshape_and_actions(
        array_shape=array.shape,
        reference_shape=reference_shape,
        array_axes=array_axes,
        reference_axes=reference_axes,
        tolerance=tolerance,
        allow_rescaling=allow_rescaling,
    )
    array = array.reshape(reshape_tuple)
    array = _numpy_rescaling(array=array, actions=actions, target_shape=reference_shape)
    array = _numpy_pad(
        array=array,
        actions=actions,
        target_shape=reference_shape,
        pad_mode=pad_mode,
        constant_values=pad_values,
    )
    array = _numpy_trim(array=array, actions=actions, target_shape=reference_shape)
    return array


def dask_match_shape(
    array: da.Array,
    reference_shape: tuple[int, ...],
    array_axes: Sequence[str],
    reference_axes: Sequence[str],
    tolerance: int = 1,
    pad_mode: str = "constant",
    pad_values: int | float = 0,
    allow_rescaling: bool = True,
) -> da.Array:
    """Match the shape of a dask array to a reference shape.

    This function will reshape, pad, trim and broadcast the input array
    to match the reference shape. If the shapes cannot be matched within
    the specified tolerance, an error is raised.

    The reference axes must be a superset of the array axes, and the order
    of the axes must be the same.

    Args:
        array (da.Array): The input array to be reshaped.
        reference_shape (tuple[int, ...]): The target shape to match.
        array_axes (Sequence[str]): The axes names of the input array.
        reference_axes (Sequence[str]): The axes names of the reference shape.
        tolerance (int): The maximum number of pixels by which dimensions
            can differ when matching shapes.
        pad_mode (str): The mode to use for padding. See numpy.pad for options.
        pad_values (int | float): The constant value to use for padding if
            pad_mode is 'constant'.
        allow_rescaling (bool): If True, when the array differs more than the
            tolerance, it will be rescalingd to the reference shape. If False,
            an error will be raised.
    """
    array_shape = tuple(int(s) for s in array.shape)
    _check_axes(
        array_shape=array_shape,
        reference_shape=reference_shape,
        array_axes=array_axes,
        reference_axes=reference_axes,
    )
    if array_shape == reference_shape:
        # Shapes already match
        return array
    array_axes = list(array_axes)
    reference_axes = list(reference_axes)

    reshape_tuple, actions = _compute_reshape_and_actions(
        array_shape=tuple(int(s) for s in array.shape),
        reference_shape=reference_shape,
        array_axes=array_axes,
        reference_axes=reference_axes,
        tolerance=tolerance,
        allow_rescaling=allow_rescaling,
    )
    array = da.reshape(array, reshape_tuple)
    array = _dask_rescaling(array=array, actions=actions, target_shape=reference_shape)
    array = _dask_pad(
        array=array,
        actions=actions,
        target_shape=reference_shape,
        pad_mode=pad_mode,
        constant_values=pad_values,
    )
    array = _dask_trim(array=array, actions=actions, target_shape=reference_shape)
    return array



from collections.abc import Sequence

import dask.array as da
import numpy as np
import zarr
from dask.array import Array as DaskArray

from ngio.common._dimensions import Dimensions
from ngio.common._roi import Roi, RoiPixels
from ngio.io_pipes._io_pipes import (
    DaskGetter,
    DaskSetter,
    DataGetter,
    DataSetter,
    NumpyGetter,
    NumpySetter,
)
from ngio.io_pipes._io_pipes_roi import roi_to_slicing_dict
from ngio.io_pipes._match_shape import dask_match_shape, numpy_match_shape
from ngio.io_pipes._ops_slices import SlicingInputType
from ngio.io_pipes._ops_transforms import TransformProtocol
from ngio.io_pipes._zoom_transform import BaseZoomTransform

##############################################################
#
# Numpy Pipes
#
##############################################################


def _numpy_label_to_bool_mask(
    label_data: np.ndarray,
    label: int | None,
    data_shape: tuple[int, ...],
    label_axes: tuple[str, ...],
    data_axes: tuple[str, ...],
    allow_rescaling: bool = True,
) -> np.ndarray:
    """Convert label data to a boolean mask."""
    if label is not None:
        bool_mask = label_data == label
    else:
        bool_mask = label_data != 0

    bool_mask = numpy_match_shape(
        array=bool_mask,
        reference_shape=data_shape,
        array_axes=label_axes,
        reference_axes=data_axes,
        allow_rescaling=allow_rescaling,
    )
    return bool_mask


def _setup_numpy_getters(
    zarr_array: zarr.Array,
    dimensions: Dimensions,
    label_zarr_array: zarr.Array,
    label_dimensions: Dimensions,
    roi: Roi | RoiPixels,
    axes_order: Sequence[str] | None = None,
    transforms: Sequence[TransformProtocol] | None = None,
    label_transforms: Sequence[TransformProtocol] | None = None,
    slicing_dict: dict[str, SlicingInputType] | None = None,
    label_slicing_dict: dict[str, SlicingInputType] | None = None,
    allow_rescaling: bool = True,
    remove_channel_selection: bool = False,
) -> tuple[NumpyGetter, NumpyGetter, dict[str, SlicingInputType]]:
    """Prepare slice kwargs for getting a masked array."""
    slicing_dict = roi_to_slicing_dict(
        roi=roi,
        pixel_size=dimensions.pixel_size,
        slicing_dict=slicing_dict,
    )

    data_getter = NumpyGetter(
        zarr_array=zarr_array,
        dimensions=dimensions,
        axes_order=axes_order,
        transforms=transforms,
        slicing_dict=slicing_dict,
        remove_channel_selection=remove_channel_selection,
    )

    if allow_rescaling:
        _zoom_transform = BaseZoomTransform(
            input_dimensions=dimensions,
            target_dimensions=label_dimensions,
            order="nearest",
        )
        if label_transforms is None or len(label_transforms) == 0:
            label_transforms = [_zoom_transform]
        else:
            label_transforms = [_zoom_transform, *label_transforms]

    label_slicing_dict = roi_to_slicing_dict(
        roi=roi,
        pixel_size=label_dimensions.pixel_size,
        slicing_dict=label_slicing_dict,
    )
    label_data_getter = NumpyGetter(
        zarr_array=label_zarr_array,
        dimensions=label_dimensions,
        axes_order=axes_order,
        transforms=label_transforms,
        slicing_dict=label_slicing_dict,
        remove_channel_selection=True,
    )
    return data_getter, label_data_getter, slicing_dict


class NumpyGetterMasked(DataGetter[np.ndarray]):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        label_zarr_array: zarr.Array,
        label_dimensions: Dimensions,
        roi: Roi | RoiPixels,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        label_transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        label_slicing_dict: dict[str, SlicingInputType] | None = None,
        fill_value: int | float = 0,
        allow_rescaling: bool = True,
        remove_channel_selection: bool = False,
    ):
        """Prepare slice kwargs for getting a masked array."""
        data_getter, label_data_getter, slicing_dict = _setup_numpy_getters(
            zarr_array=zarr_array,
            dimensions=dimensions,
            label_zarr_array=label_zarr_array,
            label_dimensions=label_dimensions,
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            label_transforms=label_transforms,
            slicing_dict=slicing_dict,
            label_slicing_dict=label_slicing_dict,
            allow_rescaling=allow_rescaling,
            remove_channel_selection=remove_channel_selection,
        )
        self._data_getter = data_getter
        self._label_data_getter = label_data_getter

        self._label_id = roi.label
        self._fill_value = fill_value
        self._allow_rescaling = allow_rescaling
        super().__init__(
            zarr_array=zarr_array,
            slicing_ops=self._data_getter.slicing_ops,
            axes_ops=self._data_getter.axes_ops,
            transforms=self._data_getter.transforms,
            roi=roi,
        )

    @property
    def label_id(self) -> int | None:
        return self._label_id

    def get(self) -> np.ndarray:
        """Get the masked data as a numpy array."""
        data = self._data_getter()
        label_data = self._label_data_getter()

        bool_mask = _numpy_label_to_bool_mask(
            label_data=label_data,
            label=self.label_id,
            data_shape=data.shape,
            label_axes=self._label_data_getter.axes_ops.output_axes,
            data_axes=self._data_getter.axes_ops.output_axes,
            allow_rescaling=self._allow_rescaling,
        )
        if bool_mask.shape != data.shape:
            bool_mask = np.broadcast_to(bool_mask, data.shape)
        masked_data = np.where(bool_mask, data, self._fill_value)
        return masked_data


class NumpySetterMasked(DataSetter[np.ndarray]):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        label_zarr_array: zarr.Array,
        label_dimensions: Dimensions,
        roi: Roi | RoiPixels,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        label_transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        label_slicing_dict: dict[str, SlicingInputType] | None = None,
        allow_rescaling: bool = True,
        remove_channel_selection: bool = False,
    ):
        """Prepare slice kwargs for setting a masked array."""
        _data_getter, _label_data_getter, slicing_dict = _setup_numpy_getters(
            zarr_array=zarr_array,
            dimensions=dimensions,
            label_zarr_array=label_zarr_array,
            label_dimensions=label_dimensions,
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            label_transforms=label_transforms,
            slicing_dict=slicing_dict,
            label_slicing_dict=label_slicing_dict,
            allow_rescaling=allow_rescaling,
            remove_channel_selection=remove_channel_selection,
        )
        self._data_getter = _data_getter
        self._label_data_getter = _label_data_getter
        self._label_id = roi.label
        self._allow_rescaling = allow_rescaling

        self._data_setter = NumpySetter(
            zarr_array=zarr_array,
            dimensions=dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_dict,
            remove_channel_selection=remove_channel_selection,
        )
        super().__init__(
            zarr_array=zarr_array,
            slicing_ops=self._data_setter.slicing_ops,
            axes_ops=self._data_setter.axes_ops,
            transforms=self._data_setter.transforms,
            roi=roi,
        )

    @property
    def label_id(self) -> int | None:
        return self._label_id

    def set(self, patch: np.ndarray) -> None:
        data = self._data_getter()
        label_data = self._label_data_getter()

        bool_mask = _numpy_label_to_bool_mask(
            label_data=label_data,
            label=self.label_id,
            data_shape=data.shape,
            label_axes=self._label_data_getter.axes_ops.output_axes,
            data_axes=self._data_getter.axes_ops.output_axes,
            allow_rescaling=self._allow_rescaling,
        )
        if bool_mask.shape != data.shape:
            bool_mask = np.broadcast_to(bool_mask, data.shape)
        masked_patch = np.where(bool_mask, patch, data)
        self._data_setter(masked_patch)


##############################################################
#
# Dask Pipes
#
##############################################################


def _dask_label_to_bool_mask(
    label_data: DaskArray,
    label: int | None,
    data_shape: tuple[int, ...],
    label_axes: tuple[str, ...],
    data_axes: tuple[str, ...],
    allow_rescaling: bool = True,
) -> DaskArray:
    """Convert label data to a boolean mask."""
    if label is not None:
        bool_mask = label_data == label
    else:
        bool_mask = label_data != 0

    bool_mask = dask_match_shape(
        array=bool_mask,
        reference_shape=data_shape,
        array_axes=label_axes,
        reference_axes=data_axes,
        allow_rescaling=allow_rescaling,
    )
    return bool_mask


def _setup_dask_getters(
    zarr_array: zarr.Array,
    dimensions: Dimensions,
    label_zarr_array: zarr.Array,
    label_dimensions: Dimensions,
    roi: Roi | RoiPixels,
    axes_order: Sequence[str] | None = None,
    transforms: Sequence[TransformProtocol] | None = None,
    label_transforms: Sequence[TransformProtocol] | None = None,
    slicing_dict: dict[str, SlicingInputType] | None = None,
    label_slicing_dict: dict[str, SlicingInputType] | None = None,
    allow_rescaling: bool = True,
    remove_channel_selection: bool = False,
) -> tuple[DaskGetter, DaskGetter, dict[str, SlicingInputType]]:
    """Prepare slice kwargs for getting a masked array."""
    slicing_dict = roi_to_slicing_dict(
        roi=roi,
        pixel_size=dimensions.pixel_size,
        slicing_dict=slicing_dict,
    )

    data_getter = DaskGetter(
        zarr_array=zarr_array,
        dimensions=dimensions,
        axes_order=axes_order,
        transforms=transforms,
        slicing_dict=slicing_dict,
        remove_channel_selection=remove_channel_selection,
    )

    if allow_rescaling:
        _zoom_transform = BaseZoomTransform(
            input_dimensions=dimensions,
            target_dimensions=label_dimensions,
            order="nearest",
        )
        if label_transforms is None or len(label_transforms) == 0:
            label_transforms = [_zoom_transform]
        else:
            label_transforms = [_zoom_transform, *label_transforms]

    label_slicing_dict = roi_to_slicing_dict(
        roi=roi,
        pixel_size=label_dimensions.pixel_size,
        slicing_dict=label_slicing_dict,
    )
    label_data_getter = DaskGetter(
        zarr_array=label_zarr_array,
        dimensions=label_dimensions,
        axes_order=axes_order,
        transforms=label_transforms,
        slicing_dict=label_slicing_dict,
        remove_channel_selection=True,
    )
    return data_getter, label_data_getter, slicing_dict


class DaskGetterMasked(DataGetter[DaskArray]):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        label_zarr_array: zarr.Array,
        label_dimensions: Dimensions,
        roi: Roi | RoiPixels,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        label_transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        label_slicing_dict: dict[str, SlicingInputType] | None = None,
        fill_value: int | float = 0,
        allow_rescaling: bool = True,
        remove_channel_selection: bool = False,
    ):
        """Prepare slice kwargs for getting a masked array."""
        _data_getter, _label_data_getter, slicing_dict = _setup_dask_getters(
            zarr_array=zarr_array,
            dimensions=dimensions,
            label_zarr_array=label_zarr_array,
            label_dimensions=label_dimensions,
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            label_transforms=label_transforms,
            slicing_dict=slicing_dict,
            label_slicing_dict=label_slicing_dict,
            allow_rescaling=allow_rescaling,
            remove_channel_selection=remove_channel_selection,
        )
        self._data_getter = _data_getter
        self._label_data_getter = _label_data_getter
        self._label_id = roi.label
        self._fill_value = fill_value
        self._allow_rescaling = allow_rescaling
        super().__init__(
            zarr_array=zarr_array,
            slicing_ops=self._data_getter.slicing_ops,
            axes_ops=self._data_getter.axes_ops,
            transforms=self._data_getter.transforms,
            roi=roi,
        )

    @property
    def label_id(self) -> int | None:
        return self._label_id

    def get(self) -> DaskArray:
        data = self._data_getter()
        label_data = self._label_data_getter()
        data_shape = tuple(int(dim) for dim in data.shape)
        bool_mask = _dask_label_to_bool_mask(
            label_data=label_data,
            label=self.label_id,
            data_shape=data_shape,
            label_axes=self._label_data_getter.axes_ops.output_axes,
            data_axes=self._data_getter.axes_ops.output_axes,
            allow_rescaling=self._allow_rescaling,
        )
        if bool_mask.shape != data.shape:
            bool_mask = da.broadcast_to(bool_mask, data.shape)
        masked_data = da.where(bool_mask, data, self._fill_value)
        return masked_data


class DaskSetterMasked(DataSetter[DaskArray]):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        label_zarr_array: zarr.Array,
        label_dimensions: Dimensions,
        roi: Roi | RoiPixels,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        label_transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        label_slicing_dict: dict[str, SlicingInputType] | None = None,
        allow_rescaling: bool = True,
        remove_channel_selection: bool = False,
    ):
        """Prepare slice kwargs for setting a masked array."""
        _data_getter, _label_data_getter, slicing_dict = _setup_dask_getters(
            zarr_array=zarr_array,
            dimensions=dimensions,
            label_zarr_array=label_zarr_array,
            label_dimensions=label_dimensions,
            roi=roi,
            axes_order=axes_order,
            transforms=transforms,
            label_transforms=label_transforms,
            slicing_dict=slicing_dict,
            label_slicing_dict=label_slicing_dict,
            allow_rescaling=allow_rescaling,
            remove_channel_selection=remove_channel_selection,
        )
        self._data_getter = _data_getter
        self._label_data_getter = _label_data_getter

        self._label_id = roi.label
        self._allow_rescaling = allow_rescaling

        self._data_setter = DaskSetter(
            zarr_array=zarr_array,
            dimensions=dimensions,
            axes_order=axes_order,
            transforms=transforms,
            slicing_dict=slicing_dict,
            remove_channel_selection=remove_channel_selection,
        )

        super().__init__(
            zarr_array=zarr_array,
            slicing_ops=self._data_setter.slicing_ops,
            axes_ops=self._data_setter.axes_ops,
            transforms=self._data_setter.transforms,
            roi=roi,
        )

    @property
    def label_id(self) -> int | None:
        return self._label_id

    def set(self, patch: DaskArray) -> None:
        data = self._data_getter()
        label_data = self._label_data_getter()
        data_shape = tuple(int(dim) for dim in data.shape)

        bool_mask = _dask_label_to_bool_mask(
            label_data=label_data,
            label=self.label_id,
            data_shape=data_shape,
            label_axes=self._label_data_getter.axes_ops.output_axes,
            data_axes=self._data_getter.axes_ops.output_axes,
            allow_rescaling=self._allow_rescaling,
        )
        if bool_mask.shape != data.shape:
            bool_mask = da.broadcast_to(bool_mask, data.shape)
        masked_patch = da.where(bool_mask, patch, data)
        self._data_setter(masked_patch)



"""Region of interest (ROI) metadata.

These are the interfaces bwteen the ROI tables / masking ROI tables and
    the ImageLikeHandler.
"""

from typing import TypeVar
from warnings import warn

from pydantic import BaseModel, ConfigDict

from ngio.common._dimensions import Dimensions
from ngio.ome_zarr_meta.ngio_specs import DefaultSpaceUnit, PixelSize, SpaceUnits
from ngio.utils import NgioValueError


def _world_to_raster(value: float, pixel_size: float, eps: float = 1e-6) -> float:
    raster_value = value / pixel_size

    # If the value is very close to an integer, round it
    # This ensures that we don't have floating point precision issues
    # When loading ROIs that were originally defined in pixel coordinates
    _rounded = round(raster_value)
    if abs(_rounded - raster_value) < eps:
        return _rounded
    return raster_value


def _to_raster(value: float, length: float, pixel_size: float) -> tuple[float, float]:
    """Convert to raster coordinates."""
    raster_value = _world_to_raster(value, pixel_size)
    raster_length = _world_to_raster(length, pixel_size)
    return raster_value, raster_length


def _to_slice(start: float | None, length: float | None) -> slice:
    if length is not None:
        assert start is not None
        end = start + length
    else:
        end = None
    return slice(start, end)


def _raster_to_world(value: int | float, pixel_size: float) -> float:
    """Convert to world coordinates."""
    return value * pixel_size


T = TypeVar("T", int, float)


class GenericRoi(BaseModel):
    """A generic Region of Interest (ROI) model."""

    name: str | None = None
    x: float
    y: float
    z: float | None = None
    t: float | None = None
    x_length: float
    y_length: float
    z_length: float | None = None
    t_length: float | None = None
    label: int | None = None
    unit: SpaceUnits | str | None = None

    model_config = ConfigDict(extra="allow")

    def intersection(self, other: "GenericRoi") -> "GenericRoi | None":
        """Calculate the intersection of this ROI with another ROI."""
        return roi_intersection(self, other)

    def _nice_str(self) -> str:
        if self.t is not None:
            t_start = self.t
        else:
            t_start = None
        if self.t_length is not None and t_start is not None:
            t_end = t_start + self.t_length
        else:
            t_end = None

        t_str = f"t={t_start}->{t_end}"

        if self.z is not None:
            z_start = self.z
        else:
            z_start = None
        if self.z_length is not None and z_start is not None:
            z_end = z_start + self.z_length
        else:
            z_end = None
        z_str = f"z={z_start}->{z_end}"

        y_str = f"y={self.y}->{self.y + self.y_length}"
        x_str = f"x={self.x}->{self.x + self.x_length}"

        if self.label is not None:
            label_str = f", label={self.label}"
        else:
            label_str = ""
        cls_name = self.__class__.__name__
        return f"{cls_name}({t_str}, {z_str}, {y_str}, {x_str}{label_str})"

    def get_name(self) -> str:
        """Get the name of the ROI, or a default if not set."""
        if self.name is not None:
            return self.name
        return self._nice_str()

    def __repr__(self) -> str:
        return self._nice_str()

    def __str__(self) -> str:
        return self._nice_str()

    def to_slicing_dict(self, pixel_size: PixelSize) -> dict[str, slice]:
        raise NotImplementedError


def _1d_intersection(
    a: T | None, a_length: T | None, b: T | None, b_length: T | None
) -> tuple[T | None, T | None]:
    """Calculate the intersection of two 1D intervals."""
    if a is None:
        if b is not None and b_length is not None:
            return b, b_length
        return None, None
    if b is None:
        if a is not None and a_length is not None:
            return a, a_length
        return None, None

    assert (
        a is not None
        and a_length is not None
        and b is not None
        and b_length is not None
    )
    start = max(a, b)
    end = min(a + a_length, b + b_length)
    length = end - start

    if length <= 0:
        return None, None

    return start, length


def roi_intersection(ref_roi: GenericRoi, other_roi: GenericRoi) -> GenericRoi | None:
    """Calculate the intersection of two ROIs."""
    if (
        ref_roi.unit is not None
        and other_roi.unit is not None
        and ref_roi.unit != other_roi.unit
    ):
        raise NgioValueError(
            "Cannot calculate intersection of ROIs with different units."
        )

    x, x_length = _1d_intersection(
        ref_roi.x, ref_roi.x_length, other_roi.x, other_roi.x_length
    )
    if x is None and x_length is None:
        # No intersection
        return None
    assert x is not None and x_length is not None

    y, y_length = _1d_intersection(
        ref_roi.y, ref_roi.y_length, other_roi.y, other_roi.y_length
    )
    if y is None and y_length is None:
        # No intersection
        return None
    assert y is not None and y_length is not None

    z, z_length = _1d_intersection(
        ref_roi.z, ref_roi.z_length, other_roi.z, other_roi.z_length
    )
    t, t_length = _1d_intersection(
        ref_roi.t, ref_roi.t_length, other_roi.t, other_roi.t_length
    )

    if (z_length is not None and z_length <= 0) or (
        t_length is not None and t_length <= 0
    ):
        # No intersection
        return None

    # Find label
    if ref_roi.label is not None and other_roi.label is not None:
        if ref_roi.label != other_roi.label:
            raise NgioValueError(
                "Cannot calculate intersection of ROIs with different labels."
            )
    label = ref_roi.label or other_roi.label

    if ref_roi.name is not None and other_roi.name is not None:
        name = f"{ref_roi.name}:{other_roi.name}"
    else:
        name = ref_roi.name or other_roi.name

    cls_ref = ref_roi.__class__
    return cls_ref(
        name=name,
        x=x,
        y=y,
        z=z,
        t=t,
        x_length=x_length,
        y_length=y_length,
        z_length=z_length,
        t_length=t_length,
        unit=ref_roi.unit,
        label=label,
    )


class Roi(GenericRoi):
    x: float = 0.0
    y: float = 0.0
    unit: SpaceUnits | str | None = DefaultSpaceUnit

    def to_roi_pixels(self, pixel_size: PixelSize) -> "RoiPixels":
        """Convert to raster coordinates."""
        x, x_length = _to_raster(self.x, self.x_length, pixel_size.x)
        y, y_length = _to_raster(self.y, self.y_length, pixel_size.y)

        if self.z is None:
            z, z_length = None, None
        else:
            assert self.z_length is not None
            z, z_length = _to_raster(self.z, self.z_length, pixel_size.z)

        if self.t is None:
            t, t_length = None, None
        else:
            assert self.t_length is not None
            t, t_length = _to_raster(self.t, self.t_length, pixel_size.t)
        extra_dict = self.model_extra if self.model_extra else {}

        return RoiPixels(
            name=self.name,
            x=x,
            y=y,
            z=z,
            t=t,
            x_length=x_length,
            y_length=y_length,
            z_length=z_length,
            t_length=t_length,
            label=self.label,
            unit=self.unit,
            **extra_dict,
        )

    def to_pixel_roi(
        self, pixel_size: PixelSize, dimensions: Dimensions | None = None
    ) -> "RoiPixels":
        """Convert to raster coordinates."""
        warn(
            "to_pixel_roi is deprecated and will be removed in a future release. "
            "Use to_roi_pixels instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self.to_roi_pixels(pixel_size=pixel_size)

    def zoom(self, zoom_factor: float = 1) -> "Roi":
        """Zoom the ROI by a factor.

        Args:
            zoom_factor: The zoom factor. If the zoom factor
                is less than 1 the ROI will be zoomed in.
                If the zoom factor is greater than 1 the ROI will be zoomed out.
                If the zoom factor is 1 the ROI will not be changed.
        """
        return zoom_roi(self, zoom_factor)

    def to_slicing_dict(self, pixel_size: PixelSize) -> dict[str, slice]:
        """Convert to a slicing dictionary."""
        roi_pixels = self.to_roi_pixels(pixel_size)
        return roi_pixels.to_slicing_dict(pixel_size)


class RoiPixels(GenericRoi):
    """Region of interest (ROI) in pixel coordinates."""

    x: float = 0
    y: float = 0
    unit: SpaceUnits | str | None = None

    def to_roi(self, pixel_size: PixelSize) -> "Roi":
        """Convert to raster coordinates."""
        x = _raster_to_world(self.x, pixel_size.x)
        x_length = _raster_to_world(self.x_length, pixel_size.x)
        y = _raster_to_world(self.y, pixel_size.y)
        y_length = _raster_to_world(self.y_length, pixel_size.y)

        if self.z is None:
            z = None
        else:
            z = _raster_to_world(self.z, pixel_size.z)

        if self.z_length is None:
            z_length = None
        else:
            z_length = _raster_to_world(self.z_length, pixel_size.z)

        if self.t is None:
            t = None
        else:
            t = _raster_to_world(self.t, pixel_size.t)

        if self.t_length is None:
            t_length = None
        else:
            t_length = _raster_to_world(self.t_length, pixel_size.t)

        extra_dict = self.model_extra if self.model_extra else {}
        return Roi(
            name=self.name,
            x=x,
            y=y,
            z=z,
            t=t,
            x_length=x_length,
            y_length=y_length,
            z_length=z_length,
            t_length=t_length,
            label=self.label,
            unit=self.unit,
            **extra_dict,
        )

    def to_slicing_dict(self, pixel_size: PixelSize) -> dict[str, slice]:
        """Convert to a slicing dictionary."""
        x_slice = _to_slice(self.x, self.x_length)
        y_slice = _to_slice(self.y, self.y_length)
        z_slice = _to_slice(self.z, self.z_length)
        t_slice = _to_slice(self.t, self.t_length)
        return {
            "x": x_slice,
            "y": y_slice,
            "z": z_slice,
            "t": t_slice,
        }


def zoom_roi(roi: Roi, zoom_factor: float = 1) -> Roi:
    """Zoom the ROI by a factor.

    Args:
        roi: The ROI to zoom.
        zoom_factor: The zoom factor. If the zoom factor
            is less than 1 the ROI will be zoomed in.
            If the zoom factor is greater than 1 the ROI will be zoomed out.
            If the zoom factor is 1 the ROI will not be changed.
    """
    if zoom_factor <= 0:
        raise NgioValueError("Zoom factor must be greater than 0.")

    # the zoom factor needs to be rescaled
    # from the range [-1, inf) to [0, inf)
    zoom_factor -= 1
    diff_x = roi.x_length * zoom_factor
    diff_y = roi.y_length * zoom_factor

    new_x = max(roi.x - diff_x / 2, 0)
    new_y = max(roi.y - diff_y / 2, 0)

    new_roi = Roi(
        name=roi.name,
        x=new_x,
        y=new_y,
        z=roi.z,
        t=roi.t,
        x_length=roi.x_length + diff_x,
        y_length=roi.y_length + diff_y,
        z_length=roi.z_length,
        t_length=roi.t_length,
        label=roi.label,
        unit=roi.unit,
    )
    return new_roi



"""This file is part of NGIO, a library for working with OME-Zarr data."""

from ngio.experimental.iterators._feature import FeatureExtractorIterator
from ngio.experimental.iterators._image_processing import ImageProcessingIterator
from ngio.experimental.iterators._segmentation import (
    MaskedSegmentationIterator,
    SegmentationIterator,
)

__all__ = [
    "FeatureExtractorIterator",
    "ImageProcessingIterator",
    "MaskedSegmentationIterator",
    "SegmentationIterator",
]



from pathlib import Path

import dask.array as da
import numpy as np
import pytest
from zarr.storage import MemoryStore

from ngio import open_ome_zarr_container
from ngio.experimental.iterators import (
    FeatureExtractorIterator,
    ImageProcessingIterator,
    MaskedSegmentationIterator,
    SegmentationIterator,
)
from ngio.utils import NgioValueError


@pytest.mark.parametrize(
    "zarr_name",
    [
        "test_image_yx.zarr",
        "test_image_cyx.zarr",
        "test_image_zyx.zarr",
        "test_image_czyx.zarr",
        "test_image_c1yx.zarr",
        "test_image_tyx.zarr",
        "test_image_tcyx.zarr",
        "test_image_tzyx.zarr",
        "test_image_tczyx.zarr",
    ],
)
def test_segmentation_iterator(images_v04: dict[str, Path], zarr_name: str):
    # Base test only the API, not the actual segmentation logic
    path = images_v04[zarr_name]
    ome_zarr = open_ome_zarr_container(path)
    image = ome_zarr.get_image()
    label = ome_zarr.get_label("label")
    iterator = SegmentationIterator(image, label, channel_selection=0, axes_order="yx")
    assert iterator.__repr__().startswith("SegmentationIterator")

    iterator = iterator.by_yx()
    assert len(iterator.rois) == image.dimensions.get("t", 1) * image.dimensions.get(
        "z", 1
    )
    iterator = iterator.by_chunks()

    iterator.require_no_regions_overlap()
    if image.is_3d or image.is_time_series:
        # 3D images should overlap in chunks
        with pytest.raises(NgioValueError):
            iterator.require_no_chunks_overlap()
    else:
        # 2D or 2D+channels does not overlap in chunks
        iterator.require_no_chunks_overlap()

    for i, (img_chunk, writer) in enumerate(iterator.iter_as_numpy()):
        label_patch = np.full(shape=img_chunk.shape, fill_value=i + 1, dtype=np.uint8)
        writer(label_patch)

    iterator.map_as_numpy(lambda x: np.zeros_like(x, dtype=np.uint8))

    for i, (img_chunk, writer) in enumerate(iterator.iter_as_dask()):
        label_patch = da.full(shape=img_chunk.shape, fill_value=i + 1, dtype=np.uint8)
        writer(label_patch)

    iterator.map_as_dask(lambda x: da.zeros_like(x, dtype=np.uint8))


@pytest.mark.parametrize(
    "zarr_name",
    [
        "test_image_yx.zarr",
        "test_image_cyx.zarr",
        "test_image_zyx.zarr",
        "test_image_czyx.zarr",
        "test_image_c1yx.zarr",
        "test_image_tyx.zarr",
        "test_image_tcyx.zarr",
        "test_image_tzyx.zarr",
        "test_image_tczyx.zarr",
    ],
)
def test_masked_segmentation_iterator(images_v04: dict[str, Path], zarr_name: str):
    # Base test only the API, not the actual segmentation logic
    path = images_v04[zarr_name]
    ome_zarr = open_ome_zarr_container(path)

    masked_label = ome_zarr.derive_label("masking_label")
    masked_label.set_array(np.ones(shape=masked_label.shape, dtype=np.uint8))
    masked_label.consolidate()

    ome_zarr.add_table(
        "masking_label_ROI_table", masked_label.build_masking_roi_table()
    )

    image = ome_zarr.get_masked_image(masking_label_name="masking_label")
    label = ome_zarr.get_label("label")

    iterator = MaskedSegmentationIterator(
        image, label, channel_selection=0, axes_order="yx"
    )

    iterator = iterator.by_yx()
    for i, (img_chunk, writer) in enumerate(iterator.iter_as_numpy()):
        label_patch = np.full(shape=img_chunk.shape, fill_value=i + 1, dtype=np.uint8)
        writer(label_patch)

    iterator.map_as_numpy(lambda x: np.zeros_like(x, dtype=np.uint8))

    for i, (img_chunk, writer) in enumerate(iterator.iter_as_dask()):
        label_patch = da.full(shape=img_chunk.shape, fill_value=i + 1, dtype=np.uint8)
        writer(label_patch)

    iterator.map_as_dask(lambda x: da.zeros_like(x, dtype=np.uint8))


@pytest.mark.parametrize(
    "zarr_name",
    [
        "test_image_yx.zarr",
        "test_image_cyx.zarr",
        "test_image_zyx.zarr",
        "test_image_czyx.zarr",
        "test_image_c1yx.zarr",
        "test_image_tyx.zarr",
        "test_image_tcyx.zarr",
        "test_image_tzyx.zarr",
        "test_image_tczyx.zarr",
    ],
)
def test_img_processing_iterator(images_v04: dict[str, Path], zarr_name: str):
    # Base test only the API, not the actual segmentation logic
    path = images_v04[zarr_name]
    ome_zarr = open_ome_zarr_container(path)
    image = ome_zarr.get_image()
    t_ome_zarr = ome_zarr.derive_image(store=MemoryStore())
    t_image = t_ome_zarr.get_image()

    iterator = ImageProcessingIterator(input_image=image, output_image=t_image)

    assert len(iterator.rois) == 1
    roi_table = image.build_image_roi_table()
    iterator = iterator.product(roi_table)
    assert len(iterator.rois) == 1

    iterator = iterator.by_zyx(strict=False)
    assert len(iterator.rois) == image.dimensions.get("t", 1)

    iterator = iterator.grid(size_x=64, size_y=64)
    for img_chunk, writer in iterator.iter_as_numpy():
        label_patch = np.zeros_like(img_chunk, dtype=np.uint8)
        writer(label_patch)

    iterator.map_as_numpy(lambda x: np.zeros_like(x, dtype=np.uint8))

    for img_chunk, writer in iterator.iter_as_dask():
        label_patch = da.zeros_like(img_chunk, dtype=np.uint8)
        writer(label_patch)

    iterator.map_as_dask(lambda x: da.zeros_like(x, dtype=np.uint8))


@pytest.mark.parametrize(
    "zarr_name",
    [
        "test_image_yx.zarr",
        "test_image_cyx.zarr",
        "test_image_zyx.zarr",
        "test_image_czyx.zarr",
        "test_image_c1yx.zarr",
        "test_image_tyx.zarr",
        "test_image_tcyx.zarr",
        "test_image_tzyx.zarr",
        "test_image_tczyx.zarr",
    ],
)
def test_features_iterator(images_v04: dict[str, Path], zarr_name: str):
    # Base test only the API, not the actual segmentation logic
    path = images_v04[zarr_name]
    ome_zarr = open_ome_zarr_container(path)

    image = ome_zarr.get_image()
    label = ome_zarr.get_label(name="label")

    feat_iterator = FeatureExtractorIterator(
        input_image=image,
        input_label=label,
        channel_selection=0,
        axes_order="xy",
    )
    feat_iterator = feat_iterator.by_yx()
    for data, seg, _ in feat_iterator.iter_as_numpy():
        assert data.shape == seg.shape



"""Utilities for OME-Zarr v04 specs.

This module provides a set of classes to internally handle the metadata
of the OME-Zarr v04 specification.

For Images and Labels implements the following functionalities:
- A function to find if a dict view of the metadata is a valid OME-Zarr v04 metadata.
- A function to convert a v04 image metadata to a ngio image metadata.
- A function to convert a ngio image metadata to a v04 image metadata.
"""

from ome_zarr_models.v04.axes import Axis as AxisV04
from ome_zarr_models.v04.coordinate_transformations import VectorScale as VectorScaleV04
from ome_zarr_models.v04.coordinate_transformations import (
    VectorTranslation as VectorTranslationV04,
)
from ome_zarr_models.v04.hcs import HCSAttrs as HCSAttrsV04
from ome_zarr_models.v04.image import ImageAttrs as ImageAttrsV04
from ome_zarr_models.v04.image_label import ImageLabelAttrs as LabelAttrsV04
from ome_zarr_models.v04.multiscales import Dataset as DatasetV04
from ome_zarr_models.v04.multiscales import Multiscale as MultiscaleV04
from ome_zarr_models.v04.multiscales import ValidTransform as ValidTransformV04
from ome_zarr_models.v04.omero import Channel as ChannelV04
from ome_zarr_models.v04.omero import Omero as OmeroV04
from ome_zarr_models.v04.omero import Window as WindowV04
from pydantic import ValidationError

from ngio.ome_zarr_meta.ngio_specs import (
    AxesHandler,
    AxesSetup,
    Axis,
    AxisType,
    Channel,
    ChannelsMeta,
    ChannelVisualisation,
    Dataset,
    ImageLabelSource,
    NgioImageMeta,
    NgioLabelMeta,
    NgioPlateMeta,
    NgioWellMeta,
    default_channel_name,
)
from ngio.ome_zarr_meta.v04._custom_models import CustomWellAttrs as WellAttrsV04


def _is_v04_image_meta(metadata: dict) -> ImageAttrsV04 | ValidationError:
    """Check if the metadata is a valid OME-Zarr v04 metadata.

    Args:
        metadata (dict): The metadata to check.

    Returns:
        bool: True if the metadata is a valid OME-Zarr v04 metadata, False otherwise.
    """
    try:
        return ImageAttrsV04(**metadata)
    except ValidationError as e:
        return e


def _is_v04_label_meta(metadata: dict) -> LabelAttrsV04 | ValidationError:
    """Check if the metadata is a valid OME-Zarr v04 metadata.

    Args:
        metadata (dict): The metadata to check.

    Returns:
        bool: True if the metadata is a valid OME-Zarr v04 metadata, False otherwise.
    """
    try:
        return LabelAttrsV04(**metadata)
    except ValidationError as e:
        return e
    raise RuntimeError("Unreachable code")


def _v04_omero_to_channels(v04_omero: OmeroV04 | None) -> ChannelsMeta | None:
    if v04_omero is None:
        return None

    ngio_channels = []
    for idx, v04_channel in enumerate(v04_omero.channels):
        channel_extra = v04_channel.model_extra

        if channel_extra is None:
            channel_extra = {}

        if "label" in channel_extra:
            label = channel_extra.pop("label")
        else:
            label = default_channel_name(idx)

        if "wavelength_id" in channel_extra:
            wavelength_id = channel_extra.pop("wavelength_id")
        else:
            wavelength_id = label

        if "active" in channel_extra:
            active = channel_extra.pop("active")
        else:
            active = True

        channel_visualisation = ChannelVisualisation(
            color=v04_channel.color,
            start=v04_channel.window.start,
            end=v04_channel.window.end,
            min=v04_channel.window.min,
            max=v04_channel.window.max,
            active=active,
            **channel_extra,
        )

        ngio_channels.append(
            Channel(
                label=label,
                wavelength_id=wavelength_id,
                channel_visualisation=channel_visualisation,
            )
        )

    v04_omero_extra = v04_omero.model_extra if v04_omero.model_extra is not None else {}
    return ChannelsMeta(channels=ngio_channels, **v04_omero_extra)


def _compute_scale_translation(
    v04_transforms: ValidTransformV04,
    scale: list[float],
    translation: list[float],
) -> tuple[list[float], list[float]]:
    for v04_transform in v04_transforms:
        if isinstance(v04_transform, VectorScaleV04):
            scale = [t1 * t2 for t1, t2 in zip(scale, v04_transform.scale, strict=True)]

        elif isinstance(v04_transform, VectorTranslationV04):
            translation = [
                t1 + t2
                for t1, t2 in zip(translation, v04_transform.translation, strict=True)
            ]
        else:
            raise NotImplementedError(
                f"Coordinate transformation {v04_transform} is not supported."
            )
    return scale, translation


def _v04_to_ngio_datasets(
    v04_multiscale: MultiscaleV04,
    axes_setup: AxesSetup,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> list[Dataset]:
    """Convert a v04 multiscale to a list of ngio datasets."""
    datasets = []

    global_scale = [1.0] * len(v04_multiscale.axes)
    global_translation = [0.0] * len(v04_multiscale.axes)

    if v04_multiscale.coordinateTransformations is not None:
        global_scale, global_translation = _compute_scale_translation(
            v04_multiscale.coordinateTransformations, global_scale, global_translation
        )

    # Prepare axes handler
    axes = []
    for v04_axis in v04_multiscale.axes:
        unit = v04_axis.unit
        if unit is not None and not isinstance(unit, str):
            unit = str(unit)
        axes.append(
            Axis(
                name=str(v04_axis.name),
                axis_type=AxisType(v04_axis.type),
                # (for some reason the type is a generic JsonValue,
                # but it should be a string or None)
                unit=v04_axis.unit,  # type: ignore
            )
        )
    axes_handler = AxesHandler(
        axes=axes,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )

    for v04_dataset in v04_multiscale.datasets:
        _scale, _translation = _compute_scale_translation(
            v04_dataset.coordinateTransformations, global_scale, global_translation
        )
        datasets.append(
            Dataset(
                path=v04_dataset.path,
                axes_handler=axes_handler,
                scale=_scale,
                translation=_translation,
            )
        )
    return datasets


def v04_to_ngio_image_meta(
    metadata: dict,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> tuple[bool, NgioImageMeta | ValidationError]:
    """Convert a v04 image metadata to a ngio image metadata.

    Args:
        metadata (dict): The v04 image metadata.
        axes_setup (AxesSetup, optional): The axes setup. This is
            required to convert image with non-canonical axes names.
        allow_non_canonical_axes (bool, optional): Allow non-canonical axes.
        strict_canonical_order (bool, optional): Strict canonical order.

    Returns:
        NgioImageMeta: The ngio image metadata.
    """
    v04_image = _is_v04_image_meta(metadata)
    if isinstance(v04_image, ValidationError):
        return False, v04_image

    if len(v04_image.multiscales) > 1:
        raise NotImplementedError(
            "Multiple multiscales in a single image are not supported in ngio."
        )

    v04_muliscale = v04_image.multiscales[0]

    channels_meta = _v04_omero_to_channels(v04_image.omero)
    axes_setup = axes_setup if axes_setup is not None else AxesSetup()
    datasets = _v04_to_ngio_datasets(
        v04_muliscale,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )

    name = v04_muliscale.name
    if name is not None and not isinstance(name, str):
        name = str(name)
    return True, NgioImageMeta(
        version="0.4",
        name=name,
        datasets=datasets,
        channels=channels_meta,
    )


def v04_to_ngio_label_meta(
    metadata: dict,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> tuple[bool, NgioLabelMeta | ValidationError]:
    """Convert a v04 image metadata to a ngio image metadata.

    Args:
        metadata (dict): The v04 image metadata.
        axes_setup (AxesSetup, optional): The axes setup. This is
            required to convert image with non-canonical axes names.
        allow_non_canonical_axes (bool, optional): Allow non-canonical axes.
        strict_canonical_order (bool, optional): Strict canonical order.

    Returns:
        NgioImageMeta: The ngio image metadata.
    """
    v04_label = _is_v04_label_meta(metadata)
    if isinstance(v04_label, ValidationError):
        return False, v04_label

    if len(v04_label.multiscales) > 1:
        raise NotImplementedError(
            "Multiple multiscales in a single image are not supported in ngio."
        )

    v04_muliscale = v04_label.multiscales[0]

    axes_setup = axes_setup if axes_setup is not None else AxesSetup()
    datasets = _v04_to_ngio_datasets(
        v04_muliscale,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )

    source = v04_label.image_label.source
    if source is None:
        image_label_source = None
    else:
        source = v04_label.image_label.source
        if source is None:
            image_label_source = None
        else:
            image_label_source = source.image
        image_label_source = ImageLabelSource(
            version="0.4",
            source={"image": image_label_source},
        )
    name = v04_muliscale.name
    if name is not None and not isinstance(name, str):
        name = str(name)

    return True, NgioLabelMeta(
        version="0.4",
        name=name,
        datasets=datasets,
        image_label=image_label_source,
    )


def _ngio_to_v04_multiscale(name: str | None, datasets: list[Dataset]) -> MultiscaleV04:
    """Convert a ngio multiscale to a v04 multiscale.

    Args:
        name (str | None): The name of the multiscale.
        datasets (list[Dataset]): The ngio datasets.

    Returns:
        MultiscaleV04: The v04 multiscale.
    """
    ax_mapper = datasets[0].axes_handler
    v04_axes = []
    for axis in ax_mapper.axes:
        v04_axes.append(
            AxisV04(
                name=axis.name,
                type=axis.axis_type.value if axis.axis_type is not None else None,
                unit=axis.unit if axis.unit is not None else None,
            )
        )

    v04_datasets = []
    for dataset in datasets:
        transform = [VectorScaleV04(type="scale", scale=list(dataset._scale))]
        if sum(dataset._translation) > 0:
            transform = (
                VectorScaleV04(type="scale", scale=list(dataset._scale)),
                VectorTranslationV04(
                    type="translation", translation=list(dataset._translation)
                ),
            )
        else:
            transform = (VectorScaleV04(type="scale", scale=list(dataset._scale)),)

        v04_datasets.append(
            DatasetV04(path=dataset.path, coordinateTransformations=transform)
        )
    return MultiscaleV04(
        axes=v04_axes, datasets=tuple(v04_datasets), version="0.4", name=name
    )


def _ngio_to_v04_omero(channels: ChannelsMeta | None) -> OmeroV04 | None:
    """Convert a ngio channels to a v04 omero."""
    if channels is None:
        return None

    v04_channels = []
    for channel in channels.channels:
        _model_extra = {
            "label": channel.label,
            "wavelength_id": channel.wavelength_id,
            "active": channel.channel_visualisation.active,
        }
        if channel.channel_visualisation.model_extra is not None:
            _model_extra.update(channel.channel_visualisation.model_extra)

        v04_channels.append(
            ChannelV04(
                color=channel.channel_visualisation.valid_color,
                window=WindowV04(
                    start=channel.channel_visualisation.start,
                    end=channel.channel_visualisation.end,
                    min=channel.channel_visualisation.min,
                    max=channel.channel_visualisation.max,
                ),
                **_model_extra,
            )
        )

    _model_extra = channels.model_extra if channels.model_extra is not None else {}
    return OmeroV04(channels=v04_channels, **_model_extra)


def ngio_to_v04_image_meta(metadata: NgioImageMeta) -> dict:
    """Convert a ngio image metadata to a v04 image metadata.

    Args:
        metadata (NgioImageMeta): The ngio image metadata.

    Returns:
        dict: The v04 image metadata.
    """
    v04_muliscale = _ngio_to_v04_multiscale(
        name=metadata.name, datasets=metadata.datasets
    )
    v04_omero = _ngio_to_v04_omero(metadata._channels_meta)

    v04_image = ImageAttrsV04(multiscales=[v04_muliscale], omero=v04_omero)
    return v04_image.model_dump(exclude_none=True, by_alias=True)


def ngio_to_v04_label_meta(metadata: NgioLabelMeta) -> dict:
    """Convert a ngio image metadata to a v04 image metadata.

    Args:
        metadata (NgioImageMeta): The ngio image metadata.

    Returns:
        dict: The v04 image metadata.
    """
    v04_muliscale = _ngio_to_v04_multiscale(
        name=metadata.name, datasets=metadata.datasets
    )
    labels_meta = {
        "multiscales": [v04_muliscale],
        "image-label": metadata.image_label.model_dump(),
    }
    v04_label = LabelAttrsV04(**labels_meta)
    return v04_label.model_dump(exclude_none=True, by_alias=True)


def v04_to_ngio_well_meta(
    metadata: dict,
) -> tuple[bool, NgioWellMeta | ValidationError]:
    """Convert a v04 well metadata to a ngio well metadata.

    Args:
        metadata (dict): The v04 well metadata.

    Returns:
        result (bool): True if the conversion was successful, False otherwise.
        ngio_well_meta (NgioWellMeta): The ngio well metadata.
    """
    try:
        v04_well = WellAttrsV04(**metadata)
    except ValidationError as e:
        return False, e

    return True, NgioWellMeta(**v04_well.model_dump())


def v04_to_ngio_plate_meta(
    metadata: dict,
) -> tuple[bool, NgioPlateMeta | ValidationError]:
    """Convert a v04 plate metadata to a ngio plate metadata.

    Args:
        metadata (dict): The v04 plate metadata.

    Returns:
        result (bool): True if the conversion was successful, False otherwise.
        ngio_plate_meta (NgioPlateMeta): The ngio plate metadata.
    """
    try:
        v04_plate = HCSAttrsV04(**metadata)
    except ValidationError as e:
        return False, e

    return True, NgioPlateMeta(**v04_plate.model_dump())


def ngio_to_v04_well_meta(metadata: NgioWellMeta) -> dict:
    """Convert a ngio well metadata to a v04 well metadata.

    Args:
        metadata (NgioWellMeta): The ngio well metadata.

    Returns:
        dict: The v04 well metadata.
    """
    v04_well = WellAttrsV04(**metadata.model_dump())
    return v04_well.model_dump(exclude_none=True, by_alias=True)


def ngio_to_v04_plate_meta(metadata: NgioPlateMeta) -> dict:
    """Convert a ngio plate metadata to a v04 plate metadata.

    Args:
        metadata (NgioPlateMeta): The ngio plate metadata.

    Returns:
        dict: The v04 plate metadata.
    """
    v04_plate = HCSAttrsV04(**metadata.model_dump())
    return v04_plate.model_dump(exclude_none=True, by_alias=True)



from pathlib import Path

import dask.array as da
import numpy as np
import pytest

from ngio import create_empty_ome_zarr, open_ome_zarr_container
from ngio.images._image import ChannelSelectionModel
from ngio.io_pipes._ops_axes import AxesOps
from ngio.io_pipes._ops_slices import SlicingOps
from ngio.utils import fractal_fsspec_store


class IdentityTransform:
    def get_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Apply the scaling transformation to a numpy array."""
        return array

    def get_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Apply the scaling transformation to a dask array."""
        return array

    def set_as_numpy_transform(
        self, array: np.ndarray, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> np.ndarray:
        """Apply the inverse scaling transformation to a numpy array."""
        return array

    def set_as_dask_transform(
        self, array: da.Array, slicing_ops: SlicingOps, axes_ops: AxesOps
    ) -> da.Array:
        """Apply the inverse scaling transformation to a dask array."""
        return array


@pytest.mark.parametrize(
    "zarr_name",
    [
        "test_image_yx.zarr",
        "test_image_cyx.zarr",
        "test_image_zyx.zarr",
        "test_image_czyx.zarr",
        "test_image_c1yx.zarr",
        "test_image_tyx.zarr",
        "test_image_tcyx.zarr",
        "test_image_tzyx.zarr",
        "test_image_tczyx.zarr",
    ],
)
def test_open_ome_zarr_container(images_v04: dict[str, Path], zarr_name: str):
    path = images_v04[zarr_name]
    ome_zarr = open_ome_zarr_container(path)

    whole_image_roi = ome_zarr.build_image_roi_table().get("image")
    image = ome_zarr.get_image()
    assert isinstance(image.__repr__(), str)
    assert image.get_roi(whole_image_roi).shape == image.shape

    label = ome_zarr.get_label("label")
    assert isinstance(label.__repr__(), str)
    roi = image.build_image_roi_table().get("image")
    image.get_roi(roi)
    label.get_roi(roi)


def test_ome_zarr_tables(cardiomyocyte_tiny_path: Path):
    cardiomyocyte_tiny_path = cardiomyocyte_tiny_path / "B" / "03" / "0"
    ome_zarr = open_ome_zarr_container(cardiomyocyte_tiny_path)

    assert isinstance(ome_zarr.__repr__(), str)
    _ = ome_zarr.images_container

    assert not ome_zarr.is_2d
    assert ome_zarr.is_3d
    assert not ome_zarr.is_time_series
    assert not ome_zarr.is_multi_channels
    assert not ome_zarr.is_2d_time_series
    assert not ome_zarr.is_3d_time_series
    assert ome_zarr.channel_labels == ["DAPI"]
    assert ome_zarr.wavelength_ids == ["A01_C01"]
    assert ome_zarr.get_channel_idx("DAPI") == 0
    assert ome_zarr.get_channel_idx(wavelength_id="A01_C01") == 0
    assert ome_zarr.num_channels == 1
    ome_zarr.set_axes_units(space_unit="micrometer")

    assert ome_zarr.list_tables() == ["FOV_ROI_table", "well_ROI_table"], (
        ome_zarr.list_tables()
    )
    assert ome_zarr.list_roi_tables() == ["FOV_ROI_table", "well_ROI_table"], (
        ome_zarr.list_roi_tables()
    )

    assert ome_zarr.list_labels() == []

    fov_roi = ome_zarr.get_table("FOV_ROI_table")
    fov_roi = ome_zarr.get_roi_table("FOV_ROI_table")
    assert len(fov_roi.rois()) == 2
    roi_table_1 = ome_zarr.get_roi_table("well_ROI_table")
    assert len(roi_table_1.rois()) == 1

    new_well_roi_table = ome_zarr.build_image_roi_table()
    ome_zarr.add_table("new_well_ROI_table", new_well_roi_table)

    assert ome_zarr.list_tables() == [
        "FOV_ROI_table",
        "well_ROI_table",
        "new_well_ROI_table",
    ]


@pytest.mark.parametrize("array_mode", ["numpy", "dask"])
def test_create_ome_zarr_container(tmp_path: Path, array_mode: str):
    # Very basic test to check if the container is working
    # to be expanded with more meaningful tests
    store = tmp_path / "ome_zarr.zarr"
    ome_zarr = create_empty_ome_zarr(
        store,
        shape=(10, 20, 30),
        chunks=(1, 20, 30),
        xy_pixelsize=0.5,
        levels=3,
        dtype="uint8",
    )

    assert isinstance(ome_zarr.__repr__(), str)
    assert ome_zarr.levels == 3
    assert ome_zarr.levels_paths == ["0", "1", "2"]
    assert ome_zarr.is_3d
    assert not ome_zarr.is_time_series
    assert not ome_zarr.is_multi_channels
    assert not ome_zarr.is_2d_time_series
    assert not ome_zarr.is_3d_time_series
    assert ome_zarr.space_unit == "micrometer"
    assert ome_zarr.time_unit is None

    ome_zarr.set_axes_units(space_unit="yoctometer", time_unit="yoctosecond")
    assert ome_zarr.space_unit == "yoctometer"
    assert ome_zarr.time_unit is None

    image = ome_zarr.get_image()

    assert image.shape == (10, 20, 30)
    assert image.dtype == "uint8"
    assert image.chunks == (1, 20, 30)
    assert image.pixel_size.x == 0.5
    assert image.meta.get_highest_resolution_dataset().path == "0"
    assert image.meta.get_lowest_resolution_dataset().path == "2"

    array = image.get_as_numpy(transforms=[IdentityTransform()])
    assert isinstance(array, np.ndarray)
    assert array.shape == (10, 20, 30)

    array_dask = image.get_as_dask(transforms=[IdentityTransform()])
    assert array_dask.shape == (10, 20, 30)

    array = image.get_array(
        x=slice(None),
        axes_order=["c", "z", "y", "x"],
        mode=array_mode,  # type: ignore
    )

    array = array + 1  # type: ignore

    image.set_array(
        array,
        x=slice(None),
        axes_order=["c", "z", "y", "x"],
        transforms=[IdentityTransform()],
    )
    image.consolidate(mode=array_mode)  # type: ignore

    # Omemeta
    ome_zarr.set_channel_meta(labels=["channel_x"])
    image = ome_zarr.get_image()
    assert image.channel_labels == ["channel_x"]
    ome_zarr.set_channel_percentiles()

    image = ome_zarr.get_image(path="2")
    assert np.mean(image.get_array()) == 1  # type: ignore

    new_ome_zarr = ome_zarr.derive_image(tmp_path / "derived2.zarr", ref_path="2")

    assert new_ome_zarr.levels == 3
    new_image = new_ome_zarr.get_image()
    assert new_image.shape == image.shape

    new_label = new_ome_zarr.derive_label("new_label")
    assert new_label.shape == image.shape
    assert new_label.meta.axes_handler.axes_names == ("z", "y", "x")

    assert new_ome_zarr.list_labels() == ["new_label"]
    assert new_ome_zarr.list_tables() == []
    assert new_ome_zarr.list_roi_tables() == []

    # Test masked image instantiation
    masked_image = new_ome_zarr.get_masked_image(masking_label_name="new_label")
    assert masked_image.shape == image.shape
    masked_label = new_ome_zarr.get_masked_label(
        label_name="new_label", masking_label_name="new_label"
    )
    assert masked_label.shape == image.shape


def test_remote_ome_zarr_container():
    url = (
        "https://raw.githubusercontent.com/"
        "fractal-analytics-platform/fractal-ome-zarr-examples/"
        "refs/heads/main/v04/"
        "20200812-CardiomyocyteDifferentiation14-Cycle1_B_03_mip.zarr/"
    )

    store = fractal_fsspec_store(url)
    ome_zarr = open_ome_zarr_container(store)

    assert ome_zarr.list_labels() == ["nuclei"]
    # assert ome_zarr.list_tables() == [
    #    "FOV_ROI_table",
    #    "nuclei_ROI_table",
    #    "well_ROI_table",
    #    "regionprops_DAPI",
    # ]

    _ = ome_zarr.get_label("nuclei", path="0")
    _ = ome_zarr.get_table("well_ROI_table")


def test_get_and_squeeze(tmp_path: Path):
    # Very basic test to check if the container is working
    # to be expanded with more meaningful tests
    store = tmp_path / "ome_zarr.zarr"
    ome_zarr = create_empty_ome_zarr(
        store,
        shape=(1, 20, 30),
        xy_pixelsize=0.5,
        levels=1,
        axes_names=["c", "y", "x"],
        dtype="uint8",
    )
    image = ome_zarr.get_image()
    assert image.shape == (1, 20, 30)
    assert image.get_array(axes_order=["c", "y", "x"]).shape == (1, 20, 30)
    image.set_array(
        np.ones((1, 20, 30), dtype="uint8"),
        axes_order=["c", "y", "x"],
    )
    assert image.get_array(axes_order=["y", "x"]).shape == (20, 30)
    image.set_array(
        np.ones((20, 30), dtype="uint8"),
        axes_order=["y", "x"],
    )
    assert image.get_array(axes_order=["y", "x"]).shape == (20, 30)
    image.set_array(
        np.ones((20, 30), dtype="uint8"),
        axes_order=["y", "x"],
    )
    assert image.get_array(axes_order=["x"], y=0, c=0).shape == (30,)
    image.set_array(
        np.ones((30,), dtype="uint8"),
        axes_order=["x"],
        y=0,
        c=0,
    )

    assert image.get_array(channel_selection=0, axes_order=["c", "y", "x"]).shape == (
        1,
        20,
        30,
    )
    assert image.get_array(channel_selection=(0,), axes_order=["y", "x"]).shape == (
        20,
        30,
    )
    assert image.get_array(channel_selection=(0,), axes_order=None).shape == (1, 20, 30)
    assert image.get_array(channel_selection=0, axes_order=None).shape == (20, 30)

    # Reordering axes and adding a virtual axis
    assert image.get_array(
        channel_selection=0, axes_order=["c", "x", "y", "virtual"]
    ).shape == (
        1,
        30,
        20,
        1,
    )

    # Test channel_labels
    image.get_as_numpy(channel_selection="channel_0")
    image.get_as_dask(channel_selection="channel_0")
    image.get_as_dask(
        channel_selection=ChannelSelectionModel(identifier="channel_0", mode="label")
    )
    image.get_as_dask(
        channel_selection=ChannelSelectionModel(identifier="0", mode="index")
    )


def test_derive_image_and_labels(tmp_path: Path):
    # Testing for #116
    store = tmp_path / "ome_zarr.zarr"
    ome_zarr = create_empty_ome_zarr(
        store,
        shape=(3, 20, 30),
        xy_pixelsize=0.5,
        levels=1,
        axes_names=["c", "y", "x"],
        dtype="uint8",
    )
    derived_ome_zarr = ome_zarr.derive_image(tmp_path / "derived.zarr")
    _ = derived_ome_zarr.derive_label("derived_label")
